=== added directory 'storm/zope'
=== added file 'storm/zope/__init__.py'
--- storm/zope/__init__.py	1970-01-01 00:00:00 +0000
+++ storm/zope/__init__.py	2007-07-17 04:42:52 +0000
@@ -0,0 +1,12 @@
+from zope.security.checker import NoProxy, BasicTypes, _available_by_default
+
+from storm.info import ObjectInfo
+
+
+# The followig is required for storm.info.get_obj_info() to have
+# access to a proxied object which is already in the store (IOW, has
+# the object info set already).  With this, Storm is able to
+# gracefully handle situations when a proxied object is passed to a
+# Store.
+_available_by_default.append("__object_info")
+BasicTypes[ObjectInfo] = NoProxy

=== added file 'storm/zope/configure.zcml'
--- storm/zope/configure.zcml	1970-01-01 00:00:00 +0000
+++ storm/zope/configure.zcml	2007-07-17 04:42:52 +0000
@@ -0,0 +1,31 @@
+<configure
+    xmlns="http://namespaces.zope.org/zope"
+    xmlns:i18n="http://namespaces.zope.org/i18n"
+    i18n_domain="zope"
+    >
+
+    <utility
+    	component=".zstorm.global_zstorm"
+        provides=".interfaces.IZStorm"
+        />
+
+    <class class="storm.store.ResultSet">
+        <allow attributes="copy config __iter__ __getitem__ any first last one
+                           order_by count max min avg sum values cached" />
+    </class>
+
+    <class class="storm.store.EmptyResultSet">
+        <allow attributes="copy config __iter__ __getitem__ any first last one
+                           order_by count max min avg sum values cached" />
+    </class>
+
+    <class class="storm.references.BoundReferenceSet">
+        <allow attributes="__iter__ find first last one any values order_by
+                           count __contains__" />
+    </class>
+
+    <class class="storm.references.BoundIndirectReferenceSet">
+        <require like_class="storm.references.BoundReferenceSet" />
+    </class>
+
+</configure>

=== added file 'storm/zope/interfaces.py'
--- storm/zope/interfaces.py	1970-01-01 00:00:00 +0000
+++ storm/zope/interfaces.py	2007-07-17 04:42:52 +0000
@@ -0,0 +1,9 @@
+from zope.interface import Interface
+
+
+class ZStormError(Exception):
+    pass
+
+
+class IZStorm(Interface):
+    pass

=== added file 'storm/zope/meta.zcml'
--- storm/zope/meta.zcml	1970-01-01 00:00:00 +0000
+++ storm/zope/meta.zcml	2007-07-17 04:42:52 +0000
@@ -0,0 +1,10 @@
+<configure xmlns:meta="http://namespaces.zope.org/meta">
+
+  <meta:directive
+        namespace="http://namespaces.zope.org/zope"
+        name="store"
+        schema=".metadirectives.IStoreDirective"
+        handler=".metaconfigure.store"
+        />
+
+</configure>

=== added file 'storm/zope/metaconfigure.py'
--- storm/zope/metaconfigure.py	1970-01-01 00:00:00 +0000
+++ storm/zope/metaconfigure.py	2007-07-17 04:42:52 +0000
@@ -0,0 +1,15 @@
+from zope import component
+
+from storm.zope.interfaces import IZStorm
+
+
+def set_default_uri(name, uri):
+    zstorm = component.getUtility(IZStorm)
+    zstorm.set_default_uri(name, uri)
+
+
+def store(_context, name, uri):
+    _context.action(discriminator=("store", name),
+                    callable=set_default_uri,
+                    args=(name, uri))
+

=== added file 'storm/zope/metadirectives.py'
--- storm/zope/metadirectives.py	1970-01-01 00:00:00 +0000
+++ storm/zope/metadirectives.py	2007-07-17 04:42:52 +0000
@@ -0,0 +1,8 @@
+from zope.interface import Interface
+from zope.schema import TextLine
+
+
+class IStoreDirective(Interface):
+
+    name = TextLine(title=u"Name", description=u"Store name")
+    uri = TextLine(title=u"URI", description=u"Database URI")

=== added file 'storm/zope/zstorm.py'
--- storm/zope/zstorm.py	1970-01-01 00:00:00 +0000
+++ storm/zope/zstorm.py	2007-07-17 04:42:52 +0000
@@ -0,0 +1,215 @@
+import threading
+import weakref
+
+from zope.testing.cleanup import addCleanUp
+from zope.interface import implements
+
+import transaction
+from transaction.interfaces import IDataManager, ISynchronizer
+from transaction._transaction import TransactionFailedError
+
+from storm.zope.interfaces import IZStorm, ZStormError
+from storm.database import create_database
+from storm.store import Store
+
+
+class ZStorm(object):
+
+    implements(IZStorm)
+
+    _databases = {}
+
+    def __init__(self):
+        self._local = threading.local()
+        self._default_databases = {}
+        self._default_uris = {}
+
+    def _reset(self):
+        for name, store in self.iterstores():
+            store.close()
+        self._local = threading.local()
+        self._databases.clear()
+        self._default_databases.clear()
+        self._default_uris.clear()
+
+    @property
+    def _stores(self):
+        try:
+            return self._local.stores
+        except AttributeError:
+            stores = weakref.WeakValueDictionary()
+            return self._local.__dict__.setdefault("stores", stores)
+
+    @property
+    def _named(self):
+        try:
+            return self._local.named
+        except AttributeError:
+            return self._local.__dict__.setdefault("named", {})
+
+    def _get_database(self, uri):
+        database = self._databases.get(uri)
+        if database is None:
+            return self._databases.setdefault(uri, create_database(uri))
+        return database
+
+    def set_default_uri(self, name, default_uri):
+        self._default_databases[name] = self._get_database(default_uri)
+        self._default_uris[name] = default_uri
+
+    def create(self, name, uri=None):
+        if uri is None:
+            database = self._default_databases.get(name)
+            if database is None:
+                raise ZStormError("Store named '%s' not found" % name)
+        else:
+            database = self._get_database(uri)
+        store = Store(database)
+        store.__synchronizer = StoreSynchronizer(store)
+
+        self._stores[id(store)] = store
+
+        if name is not None:
+            old_store = self._named.setdefault(name, store)
+            if old_store is not store:
+                raise ZStormError("Store named '%s' already exists" % name)
+        return store
+
+    def get(self, name, default_uri=None):
+        store = self._named.get(name)
+        if not store:
+            return self.create(name, default_uri)
+        return store
+
+    def remove(self, store):
+        """Remove the given store from ZStorm.
+
+        This removes any management of the store from ZStorm.
+
+        Notice that if the store was used inside the current
+        transaction, it's probably joined the transaction system as
+        a resource already, and thus it will commit/rollback when
+        the transaction system requests so.
+
+        This method will unlink the *synchronizer* from the transaction
+        system, so that once the current transaction is over it won't
+        link back to it in future transactions.
+        """
+        del self._stores[id(store)]
+        for name, named_store in self._named.items():
+            if store == named_store:
+                del self._named[name]
+        transaction.manager.unregisterSynch(store.__synchronizer)
+
+    def iterstores(self):
+        names = {}
+        for name, store in self._named.items():
+            names[id(store)] = name
+        for store in self._stores.values():
+            yield names.get(id(store)), store
+
+    def get_default_uris(self):
+        """
+        Return a list of name, uri tuples that are named as the default
+        databases for those names.
+        """
+        return self._default_uris.copy()
+
+
+class StoreSynchronizer(object):
+    """This class takes cares of plugging the store in new transactions.
+
+    Garbage collection should work fine, because the transaction manager
+    stores synchronizers as weak references. Even then, we give a hand
+    to the garbage collector by avoiding strong circular references
+    on the store.
+    """
+
+    implements(ISynchronizer)
+
+    def __init__(self, store):
+        data_manager = StoreDataManager(store)
+
+        self._store_ref = weakref.ref(store)
+        self._data_manager_ref = weakref.ref(data_manager)
+
+        # Join now ...
+        data_manager.join(transaction.get())
+
+        # ... and in the future.
+        transaction.manager.registerSynch(self)
+
+    def _join(self, trans):
+        # If the store is still alive and the transaction is in this thread.
+        store = self._store_ref()
+        if store and trans is transaction.get():
+            data_manager = self._data_manager_ref()
+            if data_manager is None:
+                data_manager = StoreDataManager(store)
+                self._data_manager_ref = weakref.ref(data_manager)
+            try:
+                data_manager.join(trans)
+            except TransactionFailedError:
+                # It means that an *already failed* transaction is trying
+                # to join us to notify that it is indeed failed.  We don't
+                # care about these (see the double_abort test case).
+                pass
+
+    def beforeCompletion(self, trans):
+        self._join(trans)
+
+    def afterCompletion(self, trans):
+        pass
+
+    def newTransaction(self, trans):
+        self._join(trans)
+
+
+class StoreDataManager(object):
+
+    implements(IDataManager)
+
+    transaction_manager = transaction.manager
+
+    def __init__(self, store):
+        self._store = store
+        self._trans = None
+
+    def join(self, trans):
+        if trans is not self._trans:
+            self._trans = trans
+            trans.join(self)
+
+    def abort(self, txn):
+        self._trans = None
+        self._store.rollback()
+
+    def tpc_begin(self, txn):
+        # Zope's transaction system will call tpc_begin() on all
+        # managers before calling commit, so flushing here may help
+        # in cases where there are two stores with changes, and one
+        # of them will fail.  In such cases, flushing earlier will
+        # ensure that both transactions will be rolled back, instead
+        # of one committed and one rolled back.
+        self._store.flush()
+
+    def commit(self, txn):
+        self._trans = None
+        self._store.commit()
+
+    def tpc_vote(self, txn):
+        pass
+
+    def tpc_finish(self, txn):
+        pass
+
+    def tpc_abort(self, txn):
+        pass
+
+    def sortKey(self):
+        return "store_%d" % id(self)
+
+
+global_zstorm = ZStorm()
+
+addCleanUp(global_zstorm._reset)

=== added directory 'tests/zope'
=== added file 'tests/zope/__init__.py'
=== added file 'tests/zope/zstorm.py'
--- tests/zope/zstorm.py	1970-01-01 00:00:00 +0000
+++ tests/zope/zstorm.py	2007-07-17 04:42:52 +0000
@@ -0,0 +1,161 @@
+import unittest
+import thread
+
+from zope import component
+
+import transaction
+
+from storm.exceptions import OperationalError
+from storm.locals import Store, Int
+from storm.zope.interfaces import IZStorm, ZStormError
+from storm.zope.zstorm import ZStorm
+
+
+class ZStormTest(unittest.TestCase):
+
+    def is_supported(self):
+        # FIXME This needs to do the right thing!
+        return True
+
+    def setUp(self):
+        self.zstorm = ZStorm()
+
+    def tearDown(self):
+        # Free the transaction to avoid having errors that cross
+        # test cases.
+        transaction.manager.free(transaction.get())
+
+    def test_utility(self):
+        component.provideUtility(ZStorm())
+        self.assertTrue(isinstance(component.getUtility(IZStorm), ZStorm))
+
+    def test_create(self):
+        store = self.zstorm.create(None, "sqlite:")
+        self.assertTrue(isinstance(store, Store))
+
+    def test_create_twice_unnamed(self):
+        store = self.zstorm.create(None, "sqlite:")
+        store.execute("CREATE TABLE test (id INTEGER)")
+        store.commit()
+
+        store = self.zstorm.create(None, "sqlite:")
+        self.assertRaises(OperationalError,
+                          store.execute, "SELECT * FROM test")
+
+    def test_create_twice_same_name(self):
+        store = self.zstorm.create("name", "sqlite:")
+        self.assertRaises(ZStormError, self.zstorm.create, "name", "sqlite:")
+
+    def test_create_and_get_named(self):
+        store = self.zstorm.create("name", "sqlite:")
+        self.assertTrue(self.zstorm.get("name") is store)
+
+    def test_create_and_get_named_another_thread(self):
+        store = self.zstorm.create("name", "sqlite:")
+
+        raised = []
+
+        lock = thread.allocate_lock()
+        lock.acquire()
+        def f():
+            try:
+                try:
+                    self.zstorm.get("name")
+                except ZStormError:
+                    raised.append(True)
+            finally:
+                lock.release()
+        thread.start_new_thread(f, ())
+        lock.acquire()
+
+        self.assertTrue(raised)
+
+    def test_get_unexistent(self):
+        self.assertRaises(ZStormError, self.zstorm.get, "name")
+
+    def test_get_with_uri(self):
+        store = self.zstorm.get("name", "sqlite:")
+        self.assertTrue(isinstance(store, Store))
+        self.assertTrue(self.zstorm.get("name") is store)
+        self.assertTrue(self.zstorm.get("name", "sqlite:") is store)
+
+    def test_set_default_uri(self):
+        self.zstorm.set_default_uri("name", "sqlite:")
+        store = self.zstorm.get("name")
+        self.assertTrue(isinstance(store, Store))
+
+    def test_create_default(self):
+        self.zstorm.set_default_uri("name", "sqlite:")
+        store = self.zstorm.create("name")
+        self.assertTrue(isinstance(store, Store))
+
+    def test_create_default_twice(self):
+        self.zstorm.set_default_uri("name", "sqlite:")
+        self.zstorm.create("name")
+        self.assertRaises(ZStormError, self.zstorm.create, "name")
+
+    def test_iterstores(self):
+        store1 = self.zstorm.create(None, "sqlite:")
+        store2 = self.zstorm.create(None, "sqlite:")
+        store3 = self.zstorm.create("name", "sqlite:")
+        stores = []
+        for name, store in self.zstorm.iterstores():
+            stores.append((name, store))
+        self.assertEquals(len(stores), 3)
+        self.assertEquals(set(stores),
+                          set([(None, store1), (None, store2),
+                               ("name", store3)]))
+
+    def test_default_databases(self):
+        self.zstorm.set_default_uri("name1", "sqlite:1")
+        self.zstorm.set_default_uri("name2", "sqlite:2")
+        self.zstorm.set_default_uri("name3", "sqlite:3")
+        default_uris = self.zstorm.get_default_uris()
+        self.assertEquals(default_uris, {"name1": "sqlite:1",
+                                         "name2": "sqlite:2",
+                                         "name3": "sqlite:3"})
+
+    def test_remove(self):
+        removed_store = self.zstorm.get("name", "sqlite:")
+        self.zstorm.remove(removed_store)
+        for name, store in self.zstorm.iterstores():
+            self.assertNotEquals(store, removed_store)
+        self.assertRaises(ZStormError, self.zstorm.get, "name")
+
+        # Abort the transaction so that the currently registered
+        # resource detaches.
+        transaction.abort()
+
+        # Let's try to make storm blow up when committing, so that
+        # we can be sure that the removed store isn't linked to the
+        # transaction system by a synchronizer anymore.
+        class BadTable(object):
+            __storm_table__ = "bad_table"
+            id = Int(primary=True)
+        removed_store.add(BadTable())
+
+        # If this fails, the store is still linked to the transaction
+        # system.
+        transaction.commit()
+
+    def test_double_abort(self):
+        """
+        Surprisingly, transaction.abort() won't detach the internal
+        transaction from the manager.  This means that when
+        transaction.begin() runs, it will detect that there's still
+        a Transaction instance around, and will call abort on it once
+        more before instantiating a new Transaction.  The Transaction's
+        abort() method will then call beforeCompletion() on our
+        synchronizer, which then tries to join() on the current
+        transaction, which is still the old one and blows up saying
+        that a previous transaction has failed (we know it, since we
+        *aborted* it).
+        """
+        class BadTable(object):
+            __storm_table__ = "bad_table"
+            id = Int(primary=True, default=1)
+        store = self.zstorm.get("name", "sqlite:")
+        store.add(BadTable())
+        self.assertRaises(OperationalError, transaction.commit)
+        transaction.abort()
+        transaction.abort()

